#!/usr/bin/python
#
# a chain of responsibilty class for stacking package sources
# for up2date
# Copyright (c) 1999-2001 Red Hat, Inc.  Distributed under GPL.
#
# Author: Adrian Likins <alikins@redhat.com>
#
"""A chain of responsibility class for stacking package sources for up2date"""

import up2dateUtils
import glob
import socket
import xmlrpclib, cgiwrap
import re
import os
import rpm
import string
import time

from translate import _, N_, cat
import rhnHTTPlib

import up2date

# Exceptions
class PartialReadError(Exception):
    pass

BUFFER_SIZE = 8092

DEBUG=0
def DPRINT(string):
    if DEBUG:
        print  string

def pkgToString(pkg):
    return "%s-%s-%s" % (pkg[0], pkg[1], pkg[2])

def pkgToStringArch(pkg):
    return "%s-%s-%s.%s" % (pkg[0], pkg[1], pkg[2], pkg[4])

def factory(aClass, *args, **kwargs):
    return apply(aClass, args, kwargs)

# this is going to be a factory. More than likely, it's input for
# it's init is going to be read from a config file. 
#
#  goals... easy "stacking" of data sources with some sort of priority
#      make the goal oblivous to the "names" of the data sources
# 
class PackageSourceChain:
    def __init__(self, headerCacheObject = None, metainfolist = None):
        self.metainfo = {}
        # lame, need to keep a list to keep the order
        self.source_list = []
        if metainfolist != None:
            for source in metainfolist:
                self.addSourceClass(source)

        self.headerCache = headerCacheObject


    def addSourceClass(self, metainfo):
        source = metainfo
        name = source['name']
        DPRINT("class name: %s" % name)
        self.source_list.append(name)
        self.metainfo[name] = factory(source['class'],
                                      source['args'],
                                      source['kargs'])

        # automagically associated the header cache for the stack with each
        # object in the chain so we can populate the cache if need be
        self.metainfo[name].addHeaderCacheObject(self.headerCache)
        DPRINT("class: %s args: %s kargs %s" % (source['class'], source['args'], source['kargs']))


    def addSourceInstance(self, metainfo):
        source = metainfo
        name = source['name']
        DPRINT("class name: %s" % name)
        self.source_list.append(name)
        self.metainfo[name] = source['object']

        # automagically associated the header cache for the stack with each
        # object in the chain so we can populate the cache if need be
        self.metainfo[name].addHeaderCacheObject(self.headerCache)
        self.metainfo[name]['name'] =  name
        
        DPRINT("object: %s name: %s" % (source['object'], name))

    def getPackage(self, pkg, MsgCallback = None, progressCallback = None):
        DPRINT(len(pkg))
        for source_key in self.source_list:
            source = self.metainfo[source_key]
            # FIXME: wrap in approriate exceptions
            package = source.getPackage(pkg, MsgCallback, progressCallback)
            if package != None:
                DPRINT("\n\n*** Package Fetched via : %s *** \n\n" % source['name'])
                #self.fetchType[pkg] = source['name']
                return package
        return None

    def getPackageSource(self, channel, pkg, MsgCallback = None, progressCallback = None):
        DPRINT(len(pkg))
        for source_key in self.source_list:
            source = self.metainfo[source_key]
            DPRINT(source)
            DPRINT(dir(source))
            # FIXME: wrap in approriate exceptions
            package = source.getPackageSource(channel, pkg, MsgCallback, progressCallback)
            if package != None:
                DPRINT("\n\n*** Package Fetched via : %s *** \n\n" % source['name'])
                return package
        return None
    

    def getHeader(self, pkg, msgCallback = None, progressCallback = None):
        # if the package source is specified, use it
        for source_key in self.source_list:
            source = self.metainfo[source_key]
            header = source.getHeader(pkg, progressCallback = progressCallback)
            # return the first one we find
            if header != None:
                DPRINT("\nHeader Fetched via: %s\n" % source['name'])
                # FIXME, the normal one only returns the header
                return (header,source['name'])
        return None

    def listPackages(self, channel, version, msgCallback = None, progressCallback = None):
        
        for source_key in self.source_list:
            source = self.metainfo[source_key]
            channelList = source.listPackages(channel, version, msgCallback, progressCallback)
            if channelList != None:
                DPRINT("\n\nlistPackages Fetched via: %s:\n" % source['name'])
                return (channelList, source['name'])
        return None
    

# template class for all the sources in the chain
class PackageSource:
    def __init__(self, cacheObject = None):
        self.headerCache = None
        self.info = {}

    def getHeader(self, pkg):
        pass

    def getPackage(self, pkg):
        pass

    def addHeaderCacheObject(self, cacheObject):
        self.headerCache = cacheObject

    def __setitem__(self, item, value):
        self.info[item] = value

    def __getitem__(self, item):
        return self.info[item]
        
class HeaderMemoryCache(PackageSource):
    def __init__(self, cacheObject = None):
        PackageSource.__init__(self,cacheObject)
    
    def getHeader(self, pkg, lite = None, msgCallback = None, progressCallback = None):
        if lite:
            if self.headerCache.has_key(pkgToStringArch(pkg), lite = 1):
                return self.headerCache.getLite(pkgToStringArch(pkg))

        if self.headerCache.has_key(pkgToStringArch(pkg)):
            return self.headerCache[pkgToStringArch(pkg)]


class LocalDisk(PackageSource):
    def __init__(self, cacheObject = None, packagePath = None):
        self.dir_list = up2dateUtils.getPackageSearchPath()
        if packagePath:
            self.dir_list = self.dir_list + packagePath
        else:
            # use the default appended package dirs
            if up2date.cfg.readEntry("packageDir"):
                self.dir_list = self.dir_list + string.split(up2date.cfg.readEntry("packageDir"),':')
        
        PackageSource.__init__(self, cacheObject = cacheObject)

    def addPackageDir(self, packagePath):
        self.dir_list = self.dir_list + packagePath

    def __saveHeader(self, hdr):
        tmp = rpm.headerLoad(hdr.unload())
        up2date.saveHeader(tmp)
        

    def getHeader(self, pkg, msgCallback = None, progressCallback = None):
        baseFileName = "%s-%s-%s.%s.rpm" % (pkg[0], pkg[1], pkg[2], pkg[4])
        for dir in self.dir_list:
            tmpFileNames = glob.glob("%s/%s" % (dir, baseFileName))
            fileName = "%s/%s" % (up2date.cfg.readEntry("storageDir"), baseFileName)

        fileNames = tmpFileNames
        if len(fileNames):
            if os.access(fileNames[0], os.R_OK):
                if not re.search("rpm$", fileNames[0]):
                    # it wasnt an rpm, so must be a header
                    if os.stat(fileNames[0])[6] == 0:
                        return None
                    fd = open(fileNames[0], "r")
                    # if this header is corrupt, rpmlib exits and we stop ;-<
                    try:
                        hdr = rpm.headerLoad(fd.read())
                    except:
                        return None
                    self.headerCache[pkgToStringArch(pkg)] = hdr
                    fd.close()
                    self.__saveHeader(hdr)
                    return hdr
                else:
                    if not up2date.checkRpmMd5(fileNames[0]):
                        fd = os.open(fileNames[0], 0)
                        try:
                            (hdr,isSource) = rpm.headerFromPackage(fd)
                        except:
                            os.close(fd)
                            raise up2date.RpmError(_("Error reading header"))
                        os.close(fd)
                        self.headerCache[pkgToStringArch(pkg)] = hdr
                        self.__saveHeader(hdr)
                        return hdr
                    else:
                        return None
            else:
                 return None        
        else:
            for dir in self.dir_list:
                fileNames = glob.glob("%s/%s.noarch.*" %
                                      (dir,
                                       pkgToString(pkg)))
            if len(fileNames):
                if os.access(fileNames[0], os.R_OK):
                    if not re.search("rpm$", fileNames[0]):
                        # it's not an rpm, must be a header
                        if os.stat(fileNames[0])[6] == 0:
                            return None
                        fd = open(fileNames[0], "r")
                        try:
                            hdr = rpm.headerLoad(fd.read())
                        except:
                            up2date.log.log_me("Corrupt header %s, skipping, will download later..." % fileNames[0])
                            return None
                        fd.close()
                        self.headerCache[pkgToStringArch(pkg)] = hdr
                        return hdr
                    else:
                        if not up2date.checkRpmMd5(fileNames[0]):
                            fd = os.open(fileNames[0], 0)
                            try:
                                (hdr,isSource) = rpm.headerFromPackage(fd)
                            except:
                                os.close(fd)
                                raise up2date.RpmError(_("Error reading header"))
                            os.close(fd)
                            up2date.log.log_me("Reading header from: %s" % fileNames)
                            self.headerCache[pkgToStringArch(pkg)] = hdr
                            return hdr
                        else:
                            return None
                else:
                    return None
            else:
                return None
    
    # this is kind of odd, since we actually just symlink to the package from
    # the cache dir, instead of copying it around, or keeping track of where
    # all the packages are from
    def getPackage(self, pkg, msgCallback = None, progressCallback = None):
        baseFileName = "%s-%s-%s.%s.rpm" % (pkg[0], pkg[1], pkg[2], pkg[4])
        for dir in self.dir_list:
            tmpFileNames = glob.glob("%s/%s" % (dir, baseFileName))
            fileName = "%s/%s" % (up2date.cfg.readEntry("storageDir"), baseFileName)
            # if the file is in the storageDir, dont symlink it to itself ;->
            if len(tmpFileNames) and not (dir == up2date.cfg.readEntry("storageDir")):
                try:
                    os.remove(fileName)
                except OSError:
                    pass
                if msgCallback:
                    msgCallback(fileName)
                os.symlink(tmpFileNames[0], fileName)
                break


class DiskCache(PackageSource):
    def __init__(self, cacheObject = None):
        self.dir_list = up2dateUtils.getPackageSearchPath()
            
        PackageSource.__init__(self, cacheObject = cacheObject)

    def __readHeaderFromRpm(self, fileNames, pkg):
         if not up2date.checkRpmMd5(fileNames[0]):
             fd = os.open(fileNames[0], 0)
             try:
                 (hdr,isSource) = rpm.headerFromPackage(fd)
             except:
                 os.close(fd)
                 raise up2date.RpmError(_("Error reading header"))
             os.close(fd)
             up2date.log.log_me("Reading header from: %s" % fileNames)
             self.headerCache[pkgToStringArch(pkg)] = hdr
             return hdr
         else:
             return None

    def __readHeaderFromFile(self, fileNames, pkg):
        if os.access(fileNames[0], os.R_OK):
            if os.stat(fileNames[0])[6] == 0:
                print "stat failed"
                return None
            fd = open(fileNames[0], "r")
            try:
                hdr = rpm.headerLoad(fd.read())
            except:
                return None
            self.headerCache[pkgToStringArch(pkg)] = hdr
            fd.close()
            return hdr
        else:
            return None
        
    def getHeader(self, pkg, msgCallback = None, progressCallback = None):
        for dir in self.dir_list:
            fileNames = glob.glob("%s/%s.%s.hdr" %
                                  (dir,
                                   pkgToString(pkg),
                                   pkg[4]) )
            # if we find anything, bail
            if len(fileNames):
                break


        if len(fileNames):
            hdr = self.__readHeaderFromFile(fileNames, pkg)
            if hdr:
                return hdr
            else:
                hdr = self.__readHeaderFromRpm(fileNames, pkg)
                return hdr

        else:
            for dir in self.dir_list:
                fileNames = glob.glob("%s/%s.noarch.hdr" %
                                      (dir,
                                       pkgToString(pkg)))

            # see if it is a .hdr file, if not try reading it as an rpm
            if len(fileNames):
                hdr = self.__readHeaderFromFile(fileNames, pkg)
                if hdr:
                    return hdr
                else:
                    hdr = self.__readHeaderFromRpm(fileNames,pkg)
                    return hdr

    def getPackage(self, pkg, msgCallback = None, progressCallback = None):

        count = 0
        baseFileName = "%s-%s-%s.%s.rpm" % (pkg[0], pkg[1], pkg[2], pkg[4])

        fileName = "%s/%s" % (up2date.cfg.readEntry("storageDir"),
                              baseFileName)
        
        # check if we already have package and that they are valid
        if os.access(fileName, os.R_OK) and not up2date.checkRpmMd5(fileName):
            if msgCallback:
                msgCallback(baseFileName)
            if progressCallback != None:
                progressCallback(1, 1)
            return 1
        else:
            return None

    def listPackages(self, channel, version, msgCallback = None, progressCallback = None):
        localFilename = "%s/%s.%s" % (up2date.cfg.readEntry("storageDir"), channel, version)
        if not os.access(localFilename, os.R_OK):
            return None

        # FIXME error handling and what not
        f = open(localFilename, "r")
        filecontents = f.read()
        # bump it full
        if progressCallback:
            progressCallback(100,100)
        return filecontents

class HttpGetSource(PackageSource):
    def __init__(self, server, proxyHost, loginInfo = None, cacheObject = None):
        self.s = server
        PackageSource.__init__(self, cacheObject = cacheObject)
        self._loginInfo=loginInfo

    def _reLoginIfNeeded(self):
        # see if the headers claim we are about to expire, if
        # so, update them
        time1 = time.time()
        up2date.log.log_me("see if we need to login again")
        newHeaders = up2date.checkExpirations(self._loginInfo)
        if newHeaders:
            up2date.log.log_me("attempting a relogin because our auth token is expired")
            self.s.set_headers(newHeaders)

    def _readFD(self, fd, size, filename, progressCallback = None):
        # Open the storage file
        f = open(filename, "w+")
        startTime = time.time()
        lastTime = startTime
        origsize = size
        while size > 0:
            curTime = time.time()
            l = min(BUFFER_SIZE, size)
            chunk = fd.read(l)
            # read guarantees that len(chunk) <= l
            l = len(chunk)
            if not l:
                # Nothing more to read
                break
            # And since the original l was smaller than size, we know 
            # now size >= 0
            size = size - l
            f.write(chunk)
            # Is it time to update the progress bar?
            if curTime - lastTime >= 1 or size == 0:
                lastTime = curTime
                bytesRead = origsize - size
                speed = bytesRead / (curTime - startTime)
                if speed != 0:
                    if origsize == 0:
                        secs = 0
                    else:
                        secs = size / speed
#                print "br: %s os: %s speed: %s secs: %s" % (bytesRead, origsize, speed, secs)
                if progressCallback != None:
                    progressCallback(bytesRead, origsize, speed, secs)

        if size > 0:
            # Probably the connection timed out
            f.close()
            os.unlink(filename)
            raise PartialReadError()
        # Flush the file to disk, then read it
        f.flush()
        # Rewind
        f.seek(0, 0)
        return f.read()
            

    def getHeader(self, package, msgCallback = None, progressCallback = None):
        hdr = None
        # package list format
        # 0        1        3       4     5     6      7
        # name, version, release, epoch, arch, size, channel
        self._reLoginIfNeeded()

        filename = "%s-%s-%s.%s.hdr" % (package[0], package[1], package[2],
            package[4])
        channel = package[6]

        filePath = "%s/%s" % (up2date.cfg.readEntry('storageDir'), filename)
        size, fd = self.s.getPackageHeader(channel, filename)

        pkgname = "%s-%s-%s" % (package[0], package[1], package[2])
        if msgCallback:
            msgCallback(pkgname)

        buffer = self._readFD(fd, size, filePath, progressCallback=progressCallback)
        fd.close()

        hdr = rpm.headerLoad(buffer)
        up2date.saveHeader(hdr)
        self.headerCache["%s-%s-%s.%s" % (hdr['name'],
                                       hdr['version'],
                                       hdr['release'],
                                       hdr['arch'])] = hdr
        return hdr

    
    def getPackage(self, package, msgCallback = None, progressCallback = None):
        self._reLoginIfNeeded()
        filename = "%s-%s-%s.%s.rpm" % (package[0], package[1], package[2],
            package[4])
        channel = package[6]
        
        filePath = "%s/%s" % (up2date.cfg.readEntry("storageDir"), filename)
        size, fd = self.s.getPackage(channel, filename)
       
        if msgCallback:
            msgCallback(filename)
            
        buffer = self._readFD(fd, size, filePath, progressCallback = progressCallback)
        fd.close()
        return buffer

    
    def getPackageSource(self, channel, package, msgCallback = None, progressCallback = None):
        self._reLoginIfNeeded()
        filename = package

        filePath = "%s/%s" % (up2date.cfg.readEntry("storageDir"), filename)
        size, fd = self.s.getPackageSource(channel, filename)

        buffer = self._readFD(fd, size, filePath, progressCallback = progressCallback)
        fd.close()
        return buffer
        

    def listPackages(self, channel, version, msgCallback = None, progressCallback = None):
        self._reLoginIfNeeded()
        filePath = "%s/%s.%s" % (up2date.cfg.readEntry("storageDir"), channel, version)
        size, fd = self.s.listPackages(channel, version)

        buffer = self._readFD(fd, size, filePath, progressCallback=progressCallback)
        fd.close()
        return buffer

# need a ton of helper functions for this one, but then, it's the main one
class Up2datePackageSource(PackageSource):
    def __init__(self, server, proxyHost, cacheObject = None):
        DPRINT(server)
        self.s = server
        id = up2date.getSystemId()
        PackageSource.__init__(self, cacheObject = cacheObject)
        

    # fetch it from the network, no caching of any sort
    def getHeader(self, pkg, lite = None, msgCallback = None, progressCallback = None):
        hdr = None

        try:
            ret = self.s.up2date.header(up2date.getSystemId(), pkg)
        except KeyboardInterrupt:
            raise up2date.CommunicationError(_("Connection aborted by the user"))
        except (socket.error, socket.sslerror), e:
            if len(e.args) > 1:
                raise up2date.CommunicationError(e.args[1])
            else:
                raise up2date.CommunicationError(e.args[0])
        except cgiwrap.ProtocolError, e:
            raise up2date.CommunicationError(e.errmsg)
        except xmlrpclib.ResponseError:
            raise up2date.CommunicationError("Broken response from the server.");
        except xmlrpclib.Fault, f:
            raise up2date.CommunicationError(f.faultString)

        bin = ret[0]
        hdr = rpm.headerLoad(bin.data)
        up2date.saveHeader(hdr)
        self.headerCache["%s-%s-%s" % (hdr['name'],
                                       hdr['version'],
                                       hdr['release'])] = hdr
        return hdr




def initHeaderSources(logininfo, serverUrl, httpServerUrl,
                      proxyUrl, proxyUser, proxyPassword):

    httpServer = rhnHTTPlib.Server(httpServerUrl, proxy = proxyUrl,
                                   username = proxyUser, password = proxyPassword,
                                   headers = logininfo)
    #create the headerCache object all the sources will use
    localHeaderCache = up2dateUtils.HeaderCache()

    ps = PackageSourceChain(headerCacheObject = localHeaderCache)

    hcs = HeaderMemoryCache(cacheObject = localHeaderCache)
    
    headerCacheSourceInfo= {'name':'memcache', 'object': hcs}
    ps.addSourceInstance(headerCacheSourceInfo)

    hldc = LocalDisk()
    ps.addSourceInstance({'name':'localdisk', 'object':hldc})

    hds = DiskCache()
    ps.addSourceInstance({'name':'diskcache', 'object':hds})

    gds = HttpGetSource(httpServer, None, loginInfo = logininfo)
    ps.addSourceInstance({'name':'get', 'object': gds})

    return ps

def callback(total, complete):
    print "-- %s bytes of %s" % (total, complete)

def updateHttpServer(packageSourceChain, logininfo, httpServerUrl, proxyUrl,
                     proxyUser, proxyPassword):

    httpServer = rhnHTTPlib.Server(httpServerUrl,
                                   proxy = proxyUrl,
                                   username = proxyUser,
                                   password = proxyPassword,
                                   headers = logininfo)


    hds = HttpGetSource(httpServer, None, loginInfo = logininfo)
    packageSourceChain.addSourceInstance({'name':'get', 'object': hds})

    return packageSourceChain
                                         

def initPackageSources(logininfo, serverUrl, httpServerUrl,
                       proxyUrl, proxyUser, proxyPassword):
    
    httpServer = rhnHTTPlib.Server(httpServerUrl, proxy = proxyUrl,
                                   username = proxyUser, password = proxyPassword,
                                   headers = logininfo)

    ps = PackageSourceChain()

    pkg_ldc = LocalDisk()
    ps.addSourceInstance({'name':'localdisk','object':pkg_ldc})

    hds = DiskCache()
    ps.addSourceInstance({'name':'diskcache', 'object':hds})

    gds = HttpGetSource(httpServer, None)
    ps.addSourceInstance({'name':'get', 'object': gds})

    return ps

def initSourcePackageSources(logininfo, serverUrl, httpServerUrl,
                             proxyUrl, proxyUser, proxyPassword):
    
    httpServer = rhnHTTPlib.Server(httpServerUrl, proxy = proxyUrl,
                                   username = proxyUser,
                                   password = proxyPassword,
                                   headers = logininfo)

    ps = PackageSourceChain()

#    hds = DiskCache()
#    ps.addSourceInstance({'name':'diskcache', 'object':hds})

    gds = HttpGetSource(httpServer, None)
    ps.addSourceInstance({'name':'get', 'object': gds})

    return ps

def initListPackageSources(logininfo, serverUrl, httpServerUrl,
                           proxyUrl, proxyUser, proxyPassword):
    
    httpServer = rhnHTTPlib.Server(httpServerUrl, proxy = proxyUrl,
                                   username = proxyUser, password = proxyPassword,
                                   headers = logininfo)
    ps = PackageSourceChain()

    hds = DiskCache()
    ps.addSourceInstance({'name':'diskcache', 'object':hds})

    gds = HttpGetSource(httpServer, None)
    ps.addSourceInstance({'name':'get', 'object':gds})

    return ps


