#!/usr/bin/python -u
#
# Client code for Update Agent
# Copyright (c) 1999-2001 Red Hat, Inc.  Distributed under GPL.
#
# Author: Preston Brown <pbrown@redhat.com>
#         Adrian Likins <alikins@redhat.com
#         Cristian Gafton <gafton@redhat.com>
#
"""Guts of the up2date client. This module provides abstration
of all the remote rpc calls that the client needs to make, includes
the dependency solve, and other assorted bits."""

import os,sys
import string
import re
import fnmatch, glob
import rpm
import socket
import cgiwrap
import xmlrpclib
import time
import StringIO
import up2dateUtils
import rpmSource
import rhnDefines
import checkbootloader
import rhnHTTPlib


from translate import _, N_, cat

import config

cfg = config.Up2dateConfig()
cfg.load()

def reloadConfig():
    cfg.load()

def version():
    # substituted to the real version by the Makefile at installation time.
    return "2.7.17"



# Where do we keep the CA certificate for RHNS?
# The servers we're talking to need to have their certs
# signed by one of these CA.
if cfg.readEntry("sslCACert"):
    rhns_ca_cert = cfg.readEntry("sslCACert")
else:
    rhns_ca_cert = "/usr/share/rhn/RHNS-CA-CERT"

DEBUG=0
def DPRINT(string):
    if DEBUG:
        print  string

# get around cached dependencies.  Gross ugly hack.
rpm.addMacro("_dbi_tags","Packages:Name:Basenames:Group:Requirename:Providename:Conflictname:Triggername")

# this is a "gloabal" variable caching the list of all available packages
global avail_package_list
avail_package_list = []


class Log:
    """attempt to log all interesting stuff, namely, anything that hits the network
       any error messages, package installs, etc"""
    def __init__(self):
        self.app = "up2date"
        if cfg.readEntry("logFile"):
            self.log_file_name = cfg.readEntry("logFile")
        else:
            self.log_file_name = "/var/log/up2date"

        self.log_info = "[%s]" % time.ctime(time.time())

    def log_me(self, *args):
        if len(args) > 0:
            s = "%s " % self.app
            for i in args:
                s = s + "%s" % i
        else:
            s = "%s " % (self.app)
        self.write_log(s)

    def trace_me(self):
        import traceback
        x = traceback.extract_stack()
        bar = string.join(traceback.format_list(x))
        self.write_log(bar)

    def write_log(self, s):
        self.log_file = open(self.log_file_name, 'a')
        msg = "%s %s\n" % (self.log_info, str(s))
        self.log_file.write(msg)
        self.log_file.flush()
        self.log_file.close()

log = Log()


class Error:
    """base class for errors"""
    def __init__(self, errmsg):
        self.errmsg = errmsg

    def __repr__(self):
        log.log_me(self.errmsg)
        return self.errmsg
    
class FileError(Error):
    """error to report when we encounter file errors (missing files/dirs,
       lack of permissions, quoat issues, etc"""
    def __repr__(self):
        msg = _("Disk error.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class RpmError(Error):
    """rpm itself raised an error condition"""
    def __repr__(self):
        msg = _("RPM error.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class PasswordError(Error):
    """Raise when the server responds with that a password is incorrect"""
    def __repr__(self):
        msg = _("Password error. The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class ConflictError(Error):
    """Raise when a rpm transaction set has a package conflict"""
    def __repr__(self):
        msg = _("RPM conflict error.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class DependencyError(Error):
    """Raise when a rpm transaction set has a dependency error"""
    def __repr__(self):
        msg = _("RPM dependency error.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class CommunicationError(Error):
    """Indicates a problem doing xml-rpc http communication with the server"""
    def __repr__(self):
        msg =  _("Error communicating with server.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class DelayError(Error):
    """Raise when the expected response from a xml-rpc call exceeds a timeout"""
    def __repr__(self):
        msg =  _("Delay error from server.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class RpmRemoveSkipListError(Error):
    """Raise when we try to remove a package on the RemoveSkipList"""
    def __repr__(self):
        msg = _("Could not remove package \"%s\". It was on the RemoveSkipList") % self.errmsg
        log.log_me(msg)
        return msg

class RpmRemoveError(Error):
    """Raise when we cant remove a package for some reason (failed deps, etc)"""
    def __init__(self, args):
        self.args = args
        self.errmsg = ""
        for key in self.args.keys():
            self.errmsg = self.errmsg + "%s failed because of %s\n" % (key, self.args[key])     
    def __repr__(self):
        return self.errmsg

       
def getSystemId():
    path = cfg.readEntry("systemIdPath")
    if not os.access(path, os.R_OK):
        return None
    
    f = open(path, "r")
    ret = f.read()
        
    f.close()
    return ret

def hasSSL():
    return hasattr(socket, "ssl")

def checkForStorageDir():
    dir = cfg.readEntry("storageDir")
    if os.access(dir, os.X_OK):
        return 0
    else:
        return 1

def getProxySetting():
    proxy = None
    proxyHost = cfg.readEntry("httpProxy")
    # legacy for backwards compat
    if proxyHost == "":
        try:
            proxyHost = cfg.readEntry("pkgProxy")
        except:
            proxyHost = None

    if proxyHost:
        if proxyHost[:7] == "http://":
            proxy = proxyHost[7:]
        else:
            proxy = proxyHost

    return proxy

def getServer(refreshCallback=None):
    global warningCallback

        
    if cfg.readEntry("enableProxy"):
        proxyHost = getProxySetting()
    else:
        proxyHost = None

    if hasSSL():
        serverUrl = cfg.readEntry("serverURL")
    else:
        serverUrl = cfg.readEntry("noSSLServerURL")

    if cfg.readEntry("enableProxyAuth"):
        if cfg.readEntry("proxyUser"):
            proxyUser = cfg.readEntry("proxyUser")
            
        if cfg.readEntry("proxyPassword"):
            proxyPassword = cfg.readEntry("proxyPassword")

    else:
        proxyUser = None
        proxyPassword = None

    lang = None
    for env in 'LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG':
        if os.environ.has_key(env):
            if not os.environ[env]:
                # sometimes unset
                continue
            lang = string.split(os.environ[env], ':')[0]
            lang = string.split(lang, '.')[0]
            break

        
    s = xmlrpclib.Server(serverUrl, proxyHost,
                         refreshCallback=refreshCallback,
                         username=proxyUser,
                         password=proxyPassword)
    if lang:
        s.setlang(lang)

    # require RHNS-CA-CERT file to be able to authenticate the SSL connections
    if not os.access(rhns_ca_cert, os.R_OK):
        if warningCallback:
            warningCallback("ERROR: can not find RHNS CA file: %s" % rhns_ca_cert)
        log.log_me("ERROR: can not find RHNS CA file: %s" % rhns_ca_cert)
        sys.exit(-1)

    # force the validation of the SSL cert
    s.use_CA_chain(rhns_ca_cert)    

    return s

def doCall(method, *args, **kwargs):
    ret = None

    attempt_count = 1
    if cfg.readEntry("networkRetries"):
        attempts = cfg.readEntry("networkRetries")
    else:
        attempts = 5
    
    while 1:
        failure = 0
        ret = None
        try:
            ret = apply(method, args, kwargs)
        except KeyboardInterrupt:
            raise CommunicationError(_("Connection aborted by the user"))
        # if we get a socket error, keep trying
        except (socket.error, socket.sslerror), e:
            log.log_me("A socket error occured: %s, attempt #%s" % (e, attempt_count))
            if attempt_count >= attempts:
                if len(e.args) > 1:
                    raise CommunicationError(e.args[1])
                else:
                    raise CommunicationError(e.args[0])
            else:
                failure = 1
        except cgiwrap.ProtocolError, e:
            log.log_me("A protocol error occured: %s , attempt #%s" % (e.errmsg, attempt_count))
            (errCode, errMsg ) = rhnHTTPlib.reportError(e.headers)
            reset = 0
            if abs(errCode) == 34:
                log.log_me("Auth token time out occured\n errmsg: %s" % errMsg)
                # this calls login, which in tern calls doCall (ie, this function)
                # but login should never get a 34, so should be safe from recursion
                updateLoginInfo()
                reset = 1
                
            if not reset:
                if attempt_count >= attempts:
                    raise CommunicationError(e.errmsg)
                else:
                    failure = 1
            
        except xmlrpclib.ResponseError:
            raise CommunicationError("Broken response from the server.");

        if failure:
            # rest for a second before trying again
            time.sleep(1)
            attempt_count = attempt_count + 1
        if ret != None:
            break
        
    return ret
    

def login():

    server = getServer()

    systemId = getSystemId()

    maybeUpdateVersion()
    log.log_me("logging into up2date server")

    loginInfo = doCall(server.up2date.login, systemId)
    
        
    log.log_me("successfully retrived authentication token from up2date server")

    DPRINT("logininfo: %s" % loginInfo)
    return loginInfo

def checkExpirations(loginInfo):
    #log.trace_me()
    skew_factor = 600
    headers = []
    if loginInfo and loginInfo.has_key(rhnDefines.AUTH_EXPIRATION):
        if time.time()+skew_factor > float(loginInfo[rhnDefines.AUTH_EXPIRATION]):
            log.log_me("Auth token has or is about to expire, acquiring a new one")
            headers = login()
        else:
            return None
    else:
        return None
    
    return headers

class ServerSettings:
    def __init__(self):
        self.xmlrpcServerUrl = cfg.readEntry("serverURL")
	if cfg.readEntry("useNoSSLForPackages"):
            self.httpServerUrl = cfg.readEntry("noSSLServerURL")
	else:
	    self.httpServerUrl = cfg.readEntry("serverURL")

        self.proxyUrl = None
        self.proxyUser = None
        self.proxyPassword = None
        
        if cfg.readEntry("enableProxy") and getProxySetting():
            self.proxyUrl = getProxySetting()
            if cfg.readEntry("enableProxyAuth"):
                if cfg.readEntry("proxyUser") and cfg.readEntry("proxyPassword"):
                    self.proxyPassword = cfg.readEntry("proxyPassword")
                    self.proxyUser = cfg.readEntry("proxyUser")
                    
    def settings(self):
        return self.xmlrpcServerUrl,self.httpServerUrl, self.proxyUrl, self.proxyUser, self.proxyPassword
    

def getUpdatedPackageSource(logininfo,headerSource,packageSource, sourcePackageSource, listPackageSource):
    settings  = ServerSettings()
    xmlrpcServerUrl, httpServerUrl, proxyUrl, proxyUser, proxyPassword = settings.settings()
    # next step would be to just pass settings to rpmSource.*.. 

    headerSource = rpmSource.updateHttpServer(headerSource, logininfo, httpServerUrl,
                                               proxyUrl, proxyUser, proxyPassword)

    packageSource = rpmSource.updateHttpServer(packageSource, logininfo, httpServerUrl,
                                               proxyUrl, proxyUser, proxyPassword)

    sourcePackageSource = rpmSource.updateHttpServer(sourcePackageSource, logininfo,
                                                     httpServerUrl, proxyUrl, proxyUser, proxyPassword)

    listPackageSource = rpmSource.updateHttpServer(listPackageSource,  logininfo, 
                                                   httpServerUrl, proxyUrl, proxyUser, proxyPassword)

    return (headerSource, packageSource,sourcePackageSource, listPackageSource)

def getPackageSource(logininfo):
    settings  = ServerSettings()
    xmlrpcServerUrl, httpServerUrl, proxyUrl, proxyUser, proxyPassword = settings.settings()

    headerSource = rpmSource.initHeaderSources(logininfo, xmlrpcServerUrl, httpServerUrl,
                                               proxyUrl, proxyUser, proxyPassword)
    packageSource = rpmSource.initPackageSources(logininfo, xmlrpcServerUrl, httpServerUrl,
                                                 proxyUrl, proxyUser, proxyPassword)
    sourcePackageSource = rpmSource.initSourcePackageSources(logininfo, xmlrpcServerUrl, httpServerUrl,
                                                             proxyUrl, proxyUser, proxyPassword)
    listPackageSource = rpmSource.initListPackageSources(logininfo, xmlrpcServerUrl, httpServerUrl,
                                                         proxyUrl, proxyUser, proxyPassword)

    return (headerSource, packageSource,sourcePackageSource, listPackageSource)

loginInfo = None
headerSource = None
packageSource = None
listPackageSource = None
sourcePackageSource = None


global loginInfo, headerSource, packageSource,sourcePackageSource, listPackageSource
def updateLoginInfo():
    log.log_me("updating login info")
    global loginInfo, headerSource, packageSource, sourcePackageSource,listPackageSource
    loginInfo = login()
    if loginInfo and headerSource and packageSource and sourcePackageSource and listPackageSource:
        # update
        headerSource, packageSource,sourcePackageSource, listPackageSource = getUpdatedPackageSource(loginInfo,
                                                                                                     headerSource,
                                                                                                     packageSource,
                                                                                                     sourcePackageSource,
                                                                                                     listPackageSource  )
    else:
        headerSource, packageSource,sourcePackageSource, listPackageSource = getPackageSource(loginInfo)
    

def pkgToString(pkg):
    return "%s-%s-%s" % (pkg[0], pkg[1], pkg[2])

def pkgToStringArch(pkg):
    return "%s-%s-%s.%s" % (pkg[0], pkg[1], pkg[2], pkg[4])


headerCache = up2dateUtils.HeaderCache()
#headerCache = {}

def getVersion():

    if cfg.readEntry("versionOverride") != "":
        return cfg.readEntry("versionOverride")
        
    db = openrpmdb()
    try:
        index = db.findbyprovides("redhat-release")[0]
    except:
        raise RpmError("Couldn't determine what version of Red Hat Linux you are running.\nIf you get this error, try running \n\n\t\trpm --rebuilddb\n\n")
    return (db[index])['version']


# mainly here to make conflicts resolution cleaner
def findDepLocal(db, dep):
    header = None
    if dep[0] == '/':
        # Treat it as a file dependency
        try:
            hdr_arry = db.findbyfile(dep)
        except:
            hdr_arry = []
            
        for n in hdr_arry:
            header = db[n]
            break
    else:
        # Try it first as a package name
        try:
            hdr_arry = db.findbyname(dep)
        except:
            hdr_arry = []
        for n in hdr_arry:
            header = db[n]
            break
        else:
            # else try it as a soname
            try:
                hdr_arry = db.findbyprovides(dep)
            except:
                hdr_arry = []
            for n in hdr_arry:
                header = db[n]
                break
            
    if header != None:
        return header
    else:
        return None

# if a user has upgraded to a newer release of Red Hat but still
# has a systemid from their older release, they need to get an updated
# systemid from the RHN servers.  This takes care of that.
def maybeUpdateVersion():
    try:
        idVer = xmlrpclib.loads(getSystemId())[0][0]['os_release']
    except:
        # they may not even have a system id yet.
        return 0

    systemVer = getVersion()
    
    if idVer != systemVer:
      s = getServer()
    
      try:
          newSystemId = doCall(s.registration.upgrade_version, getSystemId(), systemVer)
      except xmlrpclib.Fault, f:
          raise CommunicationError(f.faultString)

      path = cfg.readEntry("systemIdPath")
      dir = path[:string.rfind(path, "/")]
      if not os.access(dir, os.W_OK):
          try:
              os.mkdir(dir)
          except:
              return 0
      if not os.access(dir, os.W_OK):
          return 0

      if os.access(path, os.F_OK):
          # already have systemid file there; let's back it up
          savePath = path + ".save"
          try:
              os.rename(path, savePath)
          except:
              return 0

      f = open(path, "w")
      f.write(newSystemId)
      f.close()
      try:
          os.chmod(path, 0600)
      except:
          pass
      

def checkGPGInstallation():
    if not os.access("/usr/bin/gpg", os.X_OK):
        log.log_me("error: GPG is not installed")
        # not installed or accessible.
        return 1

    output = os.popen("/usr/bin/gpg --list-keys 2>/dev/null", "r").read()
    if not re.search("pub  1024D/DB42A60E 1999-09-23 Red Hat, Inc <security@redhat.com>", output):
        return 2

    return 0

def addGPGKey():
    # might as well log this security stuff well
    log.log_me("Adding /usr/share/rhn/RPM-GPG-KEY to keyring")
    log.log_me("/usr/bin/gpg --import /usr/share/rhn/RPM-GPG-KEY >/dev/null 2>&1")
    return os.system("/usr/bin/gpg --import /usr/share/rhn/RPM-GPG-KEY >/dev/null 2>&1")

def retrieveOnly():
    return cfg.readEntry("retrieveOnly")

def openrpmdb(option=0, dbpath=None):
    if not dbpath:
        if cfg.readEntry("dbpath"):
            dbpath = cfg.readEntry("dbpath")
        else:
            dbpath = "/var/lib/rpm/"

    rpm.addMacro("_dbpath", dbpath)

    log.log_me("Opening rpmdb in %s with option %s" % (dbpath,option))
    try:
        db = rpm.opendb(option)
    except rpm.error, e:
        raise RpmError(_("Could not open RPM database for reading.  Perhaps it is already in use?"))

    return db
        
def installedHeader(someName,db):
    h = None

    if type(someName) == type([]):
        pkgName = someName[0]
    else:
        pkgName = someName

# FIXME UGLY KLUGE FOR RPM BINDING weirdness in 7.0
    try:
	bar = db.findbyname(pkgName)
    except: 
	bar = []

    try:
        for index in bar:
            h = db[index]
            break
    except rpm.error, e:
        raise RpmError(e)

    if h != None:
        name = h['name']
        epoch = h['epoch']
        if epoch == None:
            epoch = ""
        version = h['version']
        release = h['release']
        if type(someName) == type([]):
            if (pkgName == name and
                pkgName[2] == version and
                pkgName[3] == release and
                pkgName[4] == epoch):
                return h
        else:
            if (pkgName == name):
                return h

    # not found
    return None


    
# has modified config files.  Returns a list of filenames
# Not used as of 03/01/2001. Remove?
def modifiedConfigs(header):
    fileNames = header['fileNames']
    fileFlags = header['fileflags']
    md5Sums = header['filemd5s']
    modified = []
    
    for f_i in range(len(fileNames)):
        if fileFlags[f_i] & rpm.RPMFILE_CONFIG:
            if md5Sums[f_i] != up2dateUtils.md5sum(fileNames[f_i]):
                modified.append(fileNames[f_i])

    return modified

# returns a list of config filenames from a header
# Not used as of 03/01/2001.  Remove?
def configFiles(header):
    fileNames = header['filenames']
    fileFlags = header['fileflags']
    configs = []
    
    for f_i in range(len(fileNames)):
        if fileFlags[f_i] & rpm.RPMFILE_CONFIG:
            configs.append(fileNames[f_i])

    return configs




def getAvailablePackageList(msgCallback=None, progressCallback=None):
    global avail_package_list

    # if this is non-zero, we've already fetched it
    if len(avail_package_list):
        return avail_package_list

    availPkgList = []

    try:
        log.log_me("getAvailablePackageList from network")
    except:
        pass

    channels =  loginInfo.get('X-RHN-Auth-Channels')
    DPRINT("getAvailablePackageList::channels: %s" % (channels,))
    if not channels:
        return None
    # The same package might be present in multiple channels; make sure we 
    # only keep the last one
    availPkgsDict = {}
    for channel in channels:
        # a channel maybe empty, with no version
        if not channel[1]:
            continue
        blip,type = doCall(listPackageSource.listPackages,
                           channel[0],
                           channel[1],
                           msgCallback = msgCallback,
                           progressCallback = progressCallback )
            
        tmp_args,tmp_method = xmlrpclib.loads(blip)
	for pkg in tmp_args[0]:
	    archScore = rpm.archscore(pkg[4])
            if archScore <= 0:
                # incompatible architecture
                continue

	    pkgName = pkg[0]
	    if not availPkgsDict.has_key(pkgName):
	    	availPkgsDict[pkgName] = pkg
		continue
	    
	    # Already seen this name
	    ret = up2dateUtils.comparePackages(availPkgsDict[pkgName], pkg)
	    if ret > 0:
	    	# don't care, we already have a better version
		continue
	    if ret < 0:
	    	# Better version
		availPkgsDict[pkgName] = pkg
		continue
	    # Same version, so compare the arches
            oldScore = rpm.archscore(availPkgsDict[pkgName][4])
            if oldScore <= archScore:
                continue
            # Package with a better architecture
            availPkgsDict[pkgName] = pkg

    # Sort the keys (i.e. the package names)
    avail_package_list = availPkgsDict.values()
    avail_package_list.sort()
    return avail_package_list

def getPackageSearchPath():
    dir_list = []
    dir_list.append(cfg.readEntry("storageDir"))

    dir_string = cfg.readEntry("packageDir")
    if dir_string: 
    	dir_list = dir_list + string.split(dir_string, ':')
    return dir_list


def getChannels():
    channels = loginInfo.get('X-RHN-Auth-Channels')
    return channels

# needed for actions, maybe client to at some point
def getErrataInfo(errata_id):
    s = getServer()
    packages = doCall(s.errata.getErrataInfo, getSystemId(), errata_id)
    
    return packages

# for the gui client that needs to show more info
def getChannelDetails():

    s = getServer()
    channels = doCall(s.up2date.listChannels, getSystemId())

    return channels



def subscribeChannels(channels,username,passwd):
    s = getServer()
    try:
        channels = doCall(s.up2date.subscribeChannels,
                          getSystemId(),
                          channels,
                          username,
                          passwd)
    except xmlrpclib.Fault, f:
        if f.faultCode == -36:
            raise PasswordError(f.faultString)
        else:
            raise CommunicationError(f.faultString)

def unsubscribeChannels(channels,username,passwd):
    s = getServer()
    try:
        channels = doCall(s.up2date.unsubscribeChannels,
                          getSystemId(),
                          channels,
                          username,
                          passwd)
    except xmlrpclib.Fault, f:
        if f.faultCode == -36:
            raise PasswordError(f.faultString)
        else:
            raise CommunicationError(f.faultString)

# pass the lite flag if you only want the "lite" header
def getHeader(pkg,lite=None):
    hdr = None

    hdr,type = doCall(headerSource.getHeader, pkg)
    return hdr
    
    

# implicitly places headers in the headerCache.
def getHeaders(pkgList, msgCallback = None, progressCallback = None,
               refreshCallback = None):

    headerList = []
        
    # if we get a single header, listify it
    # double check to make sure it's a list of lists and not just
    # a single list item
    if type(pkgList) == type([]):
        if len(pkgList):
            if type(pkgList[0]) == type(""):
                tmp = [pkgList]
                pkgList = tmp
        else:
            return []
    else:
        raise TypeError, "pkgList must be a list of list"


    # munge the list if we have the size include.
    # xml-rpc doesnt seem fond of that being included
    total = len(pkgList)

    # reset progress bar
    if progressCallback:
        progressCallback(0, total)

    if not refreshCallback:
        refreshCallback = 0

    for pkg in pkgList:
        if refreshCallback:
            refreshCallback()
        hdr = None
        hdr, sourcetype = doCall(headerSource.getHeader, pkg)
        
        headerList.append(hdr)

        if progressCallback:
            progressCallback(len(headerList), total)
 
    return headerList


def solveDep(unknowns,refreshCallback=None):
    retList = []
    log.log_me("solving dep for: %s" % unknowns)
    if not refreshCallback:
        refreshCallback = 0

    s = getServer(refreshCallback=refreshCallback)

    try:
        retList = doCall(s.up2date.solveDependencies, getSystemId(), unknowns)
    except xmlrpclib.Fault, f:
        if f.faultCode == -26:
            #raise RpmError(f.faultString + _(", depended on by %s") % unknowns)
            raise RpmError(f.faultString)
        else:
            raise CommunicationError(f.faultString)

    if len(retList):
        availList = getAvailablePackageList()
	availList.sort()

        availListHash = {}
        for p in availList:
            availListHash[tuple(p[:4])] = p

        newList = []
        availListNames = map(lambda p: p[0], availList)

        failedDeps = []
        for dep in retList.keys():
            if len(retList[dep]) == 0:
                failedDeps.append(dep)
                continue
            pkgs = retList[dep]
            # fixme, grab the first package that satisfies the dep
            #   but make sure we match nvre against the list of avail packages
            #   so we grab the right version of the package
            
            # in the future this should anaylize the possible solutions
            # and pick the "best" one. But since in theory, any of them
            # should work, grab the first
            pkg = None

            for p in pkgs:
                if availListHash.has_key(tuple(p[:4])):
                    pkg = p
                    break
            else:
                raise DependencyError("No package provides dependency %s" % (retList[dep],))
            if pkg[0] in availListNames:
                newList.append(availList[availListNames.index(pkg[0])])
            else:
                newList.append(pkg)
        resList = newList

    if len(failedDeps):
        message = ""
        for i in failedDeps:
            message = message + "A package providing %s could not be found.\n" % i
        raise DependencyError(message)
        
    return resList

def saveHeader(hdr):
    fileName = "%s/%s.%s.hdr" % (cfg.readEntry("storageDir"),
                              string.join((hdr['name'],
                                           hdr['version'],
                                           hdr['release']), "-"), hdr['arch'])

    fd = os.open(fileName, os.O_WRONLY|os.O_CREAT, 0600)

    os.write(fd, hdr.unload())
    os.close(fd)

    return 1

    
def removeHeader(pkg):
    # first see if it is on disk:
    fileNames = glob.glob("%s/%s.%s.hdr" %
                          (cfg.readEntry("storageDir"),
                           pkgToString(pkg),
                           up2dateUtils.archGlob()))
    for fileName in fileNames:
        try:
            os.remove(fileName)
        except:
            raise FileError(_("cannot remove %s.") % fileName)

    fileNames = glob.glob("%s/%s.noarch.hdr" %
                          (cfg.readEntry("storageDir"),
                           pkgToString(pkg)))
    for fileName in fileNames:
        try:
            os.remove(fileName)
        except:
            raise FileError(_("cannot remove %s.") % fileName)

def checkRpmMd5(fileName):
    # FIXME
    #insert magic to "shutup" rpm here 
    # pgb 03/01/2001: not sure if any magic can fix that in RPM. :(
    
    # only newer RPMs allow this
    if "checksig" in dir(rpm):
        # RPM spews to stdout/stderr.  Redirect.
        saveStdout = os.dup(1)
        saveStderr = os.dup(2)
        redirStdout = os.open("/dev/null", os.O_WRONLY | os.O_APPEND)
        redirStderr = os.open("/dev/null", os.O_WRONLY | os.O_APPEND)
        os.dup2(redirStdout, 1)
        os.dup2(redirStderr, 2)
        # now do the rpm thing
        ret = rpm.checksig(fileName, rpm.RPMTAG_SIGMD5)
        # restore normal stdout and stderr
        os.dup2(saveStdout, 1)
        os.dup2(saveStderr, 2)
        # Clean up
        os.close(redirStdout)
        os.close(redirStderr)
        os.close(saveStdout)
        os.close(saveStderr)
        return ret
    else:
        # legacy 3.0.x RPM
        sigOk = 0
        lines = os.popen("/bin/rpm -Kv --nopgp --nogpg %s 2>&1" % fileName, "r").read()
        if not re.search("MD5 sum OK:", lines):
            sigOk = 1

        return sigOk

def getPackage(pkg, msgCallback = None, progressCallback = None):

    buffer = doCall(packageSource.getPackage, pkg, msgCallback, progressCallback)

    # retrieve source packages as well
    if cfg.readEntry('retrieveSource'):
        hdr = getHeader(pkg)
        srcrpm = hdr['sourcerpm']
        channel = pkg[6]
        buffer = doCall(sourcePackageSource.getPackageSource,
                        channel,
                        srcrpm,
                        msgCallback,
                        progressCallback)
        
    return 0


def removePackage(pkg):
     # first see if it is on disk:
    fileNames = glob.glob("%s/%s.%s.*" %
                          (cfg.readEntry("storageDir"),
                           pkgToString(pkg),
                           up2dateUtils.archGlob()))
    for fileName in fileNames:
        log.log_me("deleting %s" % fileName)
        try:
            os.remove(fileName)
            
        except:
            raise FileError(_("Cannot remove %s") % fileName)
        
    fileNames = glob.glob("%s/%s.noarch.*" %
                          (cfg.readEntry("storageDir"),
                           pkgToString(pkg)))
    for fileName in fileNames:
        log.log_me("deleting %s" % fileName)
        try:
            os.remove(fileName)
        except:
            raise FileError(_("Cannot remove %s") % fileName)


def getAdvisoryInfo(pkg):
    global warningCallback
    s = getServer()

    db = openrpmdb()
    try:
        indices = db.findbyname(pkg[0])
    except:
        indices = []

    h = None
    for i in indices:
        h = db[i]

    if h == None:
        return None
    
    try:
        pkgName = "%s-%s-%s" % (h['name'],
                                h['version'],
                                h['release'])
        log.log_me("getAdvisoryInfo for %s" % pkgName)
        info = doCall(s.errata.getPackageErratum,
                      getSystemId(),
                      pkg)
    except xmlrpclib.Fault, f:
        if warningCallback:
            warningCallback(f.faultString)
        return None
    
    if info:
        return info
    
    try:
        log.log_me("getAdvisoryInfo for %s-0-0" % h['name'])
        info = doCall(s.errata.GetByPackage,
                      "%s-0-0" % h['name'],
                      getVersion())
    except xmlrpclib.Fault, f:
        if warningCallback:
            warningCallback(f.faultString)
        return None
    
    return info
    

def remoteAddPackages(pkg):
    log.log_me("Adding packages to package profile: %s" % pprint_pkglist(pkg))
    s = getServer()
    try:
        if type(pkg[0]) == type([]):
            #s.set_transport(2, 2)
            doCall(s.registration.add_packages,getSystemId(), pkg)
        else:
            doCall(s.registration.add_packages,getSystemId(), [pkg])            
    except xmlrpclib.Fault, f:
        raise CommunicationError(f.faultString)


def remoteDelPackages(pkg):
    log.log_me("Removing packages from package profile: %s" % pprint_pkglist(pkg))
    s = getServer()

    try:
        if type(pkg[0]) == type([]):
            #s.set_transport(2, 2)
            doCall(s.registration.delete_packages,getSystemId(), pkg)
        else:
            doCall(s.registration.delete_packages,getSystemId(), [pkg])            
    except xmlrpclib.Fault, f:
        raise CommunicationError(f.faultString)


def updatePackageProfile():
    log.log_me("Updating package profile")
    s = getServer()
    
    try:
        #s.set_transport(2, 2)
        doCall(s.registration.update_packages,
               getSystemId(),
               getInstalledPackageList())
    except xmlrpclib.Fault, f:
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)


def getInstalledPackageList(msgCallback = None, progressCallback = None, getArch=None):
    list = []

    if msgCallback != None:
        msgCallback(_("Getting list of packages installed on the system"))

    db = openrpmdb()

    if progressCallback != None:
        count = 1
        total = 0

        if "firstkey" in dir(db):
            key = db.firstkey()
            h = db[key]
        else:
            iterator = db.match(0)
            h = iterator.next()
        while (h):
            count = count + 1
            if "nextkey" in dir(db):
                key = db.nextkey(key)
                if key:
                    h = db[key]
                else:
                    h = None
            else:
                h = iterator.next()

        total = count
        count = 1

    if "firstkey" in dir(db):
        key = db.firstkey()
        h = db[key]
    else:
        iterator = db.match(0)
        h = iterator.next()
    while (h):
        name = h['name']
        epoch = h['epoch']
        if epoch == None:
            epoch = ""
        version = h['version']
        release = h['release']
        if getArch:
            arch = h['arch']
            list.append([name, version, release, epoch, arch])
        else:
            list.append([name, version, release, epoch])
            
        if progressCallback != None:
            progressCallback(count, total)
            count = count + 1
        if "nextkey" in dir(db):
            key = db.nextkey(key)
            if key:
                h = db[key]
            else:
                h = None
        else:
            h = iterator.next()

    list.sort()
    return list

def getArchesOfAvailablePackages(availablePkgList,pkgName):
    archList = []
    for pkg in availablePkgList:
        if pkgName == pkg[0]:
            archList.append(pkg[4])

    return archList

# convience class to make removeInstalledPackagesFromList cleaner
class UpdateList:
    def __init__(self):
        self.updateDict = {}
        self.__updateIndex = None

    def nameList(self):
        return self.updateDict.keys()

    # gen a list of just the name,version,release,epoch, and arch fields
    def nvreaList(self):
        return map(lambda a: a[:5], self.updatedList())

    def nvreList(self):
        return map(lambda a: a[:4], self.updatedList())

    def appendIfNewer(self, installedPkg, pkg):
        pkgName = pkg[0]
        # do we already have a package like this?
        if not self.updateDict.has_key(pkgName):
            self.updateDict[pkgName] = pkg
            return
        # do we have a match for name-version-release
        if pkg[:4] in self.nvreList():
            # see if this is a better matching architecture
	    archScore = rpm.archscore(pkg[4])
            # incompatible arch
            if archScore <= 0:
                return
	    oldScore = rpm.archscore(self.updateDict[pkgName][4])
            if archScore < oldScore: # retain the better arch
                self.updateDict[pkgName][4] = pkg[4]
            return
        # else, we might gave an older one that needs updating
        ret = up2dateUtils.comparePackages(self.updateDict[pkgName],pkg)
        if ret > 0: # if pkg is newer
            self.updateDict[pkgName] = pkg

    def updatedList(self):
        return self.updateDict.values()

    def delete(self, pkg):
        del self.updateDict[pkg[0]]
            

# returns new pkgList
# pkgList = available Packages to install
# pkgNames = list of packages to install regardless (commandline)
def removeInstalledPackagesFromList(pkgList, pkgNames,
                                    msgCallback = None,
                                    progressCallback = None):
    if pkgList:
        pkgCount = len(pkgList)
    else:
        return []

    availablePkgList = getAvailablePackageList()
    installedPkgList = getInstalledPackageList(getArch=1)


    if msgCallback != None:
        msgCallback(_("Removing installed packages from list of updates"))


    # pretend the pkgNames are installed, with versions 0-0:0
    # is this sane? it sure makes life easier...
    installedPkgListNames = map(lambda a: a[0], installedPkgList)
    for pkg in pkgNames:
        if not pkg[0] in installedPkgListNames:
            arches = getArchesOfAvailablePackages(availablePkgList,pkg[0])
            if len(arches) == 1:
                installedPkgList.append([pkg[0], '0.0','0.0','', arches[0]])
            else:
                installedPkgList.append([pkg[0], '0.0','0.0','',"noarch"])

    updatedList = []
    installedPkgCount = len(installedPkgList)
    installedIndex = 0
    
    multiArchPkgs = {}
    updateList = UpdateList()
    installedPkgListNVRE = map(lambda a: a[:4], installedPkgList)
    for installedPkg in installedPkgList:
        for index in range(pkgCount):
            if ( installedPkg[0] == pkgList[index][0] ):
              ret = up2dateUtils.comparePackages(installedPkg,pkgList[index])
              if ret == 0:
                  # already installed
                  # make sure it wasn't added to the list already,
                  # which can happen when multiple packages with the
                  # same name are installed.
                  try:
                      updateList.delete(pkgList[index])
                  except:
                      pass
              elif ret > 0:
                  # newer version already installed
                  # make sure it wasn't added to the list already,
                  # which can happen when multiple packages with the
                  # same name are installed.
                  try:
                      updateList.delete(pkgList[index])
                  except:
                      pass
              else:
                  # we need to fetch it. avoid including multiple copies;
                  # sometimes two packages w/the same name are installed
                  # on a system.
                  if pkgList[index][:4] not in installedPkgListNVRE:
                      updateList.appendIfNewer(installedPkg, pkgList[index])
                  
        installedIndex = installedIndex + 1
        if progressCallback:
            progressCallback(installedIndex, installedPkgCount)

    return updateList.updatedList()


        
    
# side effects on pkgList
def removeSkipPackagesFromList(pkgList, msgCallback = None,
                               progressCallback = None):
    removedList = []
    pkgCount = len(pkgList)

    if cfg.readEntry('forceInstall'):
        return removedList
    
    if msgCallback != None:
        msgCallback(_("Removing packages marked to skip from list"))

    pkgSkipList = cfg.readEntry("pkgSkipList")
    for index in range(pkgCount):
        for pattern in pkgSkipList:
            if fnmatch.fnmatch(pkgList[index][0], pattern):
                removedList.append((pkgList[index], _("Pkg name/pattern")))
                break

        if progressCallback != None:
            progressCallback(index+1, pkgCount)

    for pkg in removedList:
        pkgList.remove(pkg[0])

    # just do a side effect on pkgList here, and return removedList
    # since we actually need the results of it
    return removedList


# check to see if a new file has a different MD5 sum
# from the package already on disk, and if true, if
# the on-disk file has been modified.
# this is a lot of args, but it helps...
def checkModified(index, fileNames, fileMD5s,installedFileNames,installedFileMD5s):
    ret = 0
    fileName = fileNames[index]

    #this is a little ugly, but the order of the filelist in the old and the new
    # pacakges arent the same.
    for j in range(len(installedFileNames)):
        if (installedFileNames[j] == fileName):
            if installedFileMD5s[j] == fileMD5s[index]:
                # the md5 of the file in the local db  is the same as the one in the
                # incoming package, so skip the rest of the check and return 0
                continue
	    # grrr, symlinks marked as config files pointing to dirs are baaad, okay?
	    if not os.path.isdir(fileName):
            	if installedFileMD5s[j] != up2dateUtils.md5sum(fileName):
                    # the local file is different than the file in the local db
                    ret = 1
            	else:
                    # not changed on disk
                    pass
            break
    return ret


# check if the signature matches the redhat.com one
def hasBadSignature(pkg):
    hdr = getHeader(pkg)
    fileName = "%s/%s.%s.rpm" % (cfg.readEntry("storageDir"),
                                 pkgToString(pkg),
                                 hdr['arch'])
    # must use the actual header from the real package.  The headers
    # the server returns by themselves have no signature info in them.
    fd = os.open(fileName, os.O_RDONLY)
    try:
        hdr = rpm.headerFromPackage(fd)[0]
    except:
        raise RpmError(_("Failed to read header from %s for signature verification") % fileName)
    
    os.close(fd)

    try:
        # only newer RPMs allow this
        if hdr['SIGGPG'] == None:
            return 1
        if "checksig" in dir(rpm):
            # RPM spews to stdout/stderr.  Redirect.
            saveStdout = os.dup(1)
            saveStderr = os.dup(2)
            redirStdout = os.open("/dev/null", os.O_WRONLY | os.O_APPEND)
            redirStderr = os.open("/dev/null", os.O_WRONLY | os.O_APPEND)
            os.dup2(redirStdout, 1)
            os.dup2(redirStderr, 2)
            # now do the rpm thing
            ret = rpm.checksig(fileName, rpm.RPMTAG_SIGGPG)
            # restore normal stdout and stderr
            os.dup2(saveStdout, 1)
            os.dup2(saveStderr, 2)
            # clean up
            os.close(redirStdout)
            os.close(redirStderr)
            os.close(saveStdout)
            os.close(saveStderr)
            if ret:
                return 2

        return 0
    except KeyError:
        # legacy 3.0.x RPM
        sigOk = 1
        lines = os.popen("/bin/rpm -Kv %s 2>&1" % fileName, "r").read()
        if re.search("BAD signature from \"Red Hat, Inc <security@redhat.com>\"", lines):
            sigOk = 2
        elif re.search("Good signature from \"Red Hat, Inc <security@redhat.com>\"", lines) and re.search("Fingerprint: CA20 8686 2BD6 9DFC 65F6  ECC4 2191 80CD DB42 A60E", lines):
            sigOk = 0

        return sigOk

# trim down pkgList to just the files in pkgNames, ie the ones
# specified on the commandline. ugly function name, but it sorta
# follows the existing scheme. We really should just be requesting
# info on the file specified on the commandline, but this works for now.
#
# returns the new pkgList
def removeNonPkgNamesPackagesFromList(pkgList,pkgNames,msgCallback = None,
                                      progressCallback = None):
    
    if msgCallback != None:
        msgCallback(_("Removing packages with files not specified from list"))

    tmp_pkgList = []
    for pkg in pkgList:
        names = map(lambda a: a[0], pkgNames)
        if (pkg[0] in names):
            tmp_pkgList.append(pkg)

                
    return tmp_pkgList


def checkHeaderForFileConfigExcludes(h,package,db):
    fflag = 0

    # might as well just do this once...
    fileNames = h['filenames']
    fileMD5s = h['filemd5s']
    fileFlags = h['fileflags']

    installedHdr = installedHeader(h['name'],db)
    if not installedHdr:
        #throw a fault? should this happen?
        return None
        
    installedFileNames = installedHdr['filenames']
    installedFileMD5s = installedHdr['filemd5s']
    
    removedList = []
    if cfg.readEntry('forceInstall'):
        return None

    # bleah, have to use the index here because
    # of rpm's storing of filenames and their md5sums in parallel lists
    for f_i in range(len(fileNames)):
        # code to see if we want to disable this
        for pattern in cfg.readEntry("fileSkipList"):
            
            if fnmatch.fnmatch(fileNames[f_i],pattern):
                # got to get a better string to use here
                removedList.append((package, _("File Name/pattern")))
                fflag = 1
                break
            # if we found a matching file, no need to
            # examine the rest in this package         
            if fflag:
                break

    if cfg.readEntry("configFilesToIgnore"):
        configFilesToIgnore = cfg.readEntry("configFilesToIgnore")
    else:
        configFilesToIgnore = []

    for f_i in range(len(fileNames)):
        # Deal with config files
        if cfg.readEntry("noReplaceConfig"):
            # check for files that are config files, but skips those that
            # arent going to be replaced anyway
            # (1 << 4) == rpm.RPMFILE_NOREPLACE if it existed
            if fileFlags[f_i] & rpm.RPMFILE_CONFIG and not \
               fileFlags[f_i] & (1 << 4):
                if fileNames[f_i] not in configFilesToIgnore:
                    # check if config file and if so, if modified
                    if checkModified(f_i, fileNames, fileMD5s,
                                     installedFileNames, installedFileMD5s):
                        removedList.append((package, _("Config modified")))
                        fflag = 1
                        break

        if fflag:
            break     

    if len(removedList):
        return removedList
    else:
        return None

# side effects on pkgList,
def removeSkipFilesPackagesFromList(pkgList, msgCallback = None,
                                    progressCallback = None):
    removedList = []

    if msgCallback != None:
        msgCallback(_("Removing packages with files marked to skip from list"))

    p_i = 0
    h_list = getHeaders(pkgList, progressCallback=progressCallback)

    # need the db for the checkHeaderFiles
    db = openrpmdb()
    
    if not cfg.readEntry('forceInstall'):
        for i in h_list:
            tmp_list = 0
            #check the files based on the new header
            tmp_list = checkHeaderForFileConfigExcludes(i,pkgList[p_i],db)
            p_i = p_i + 1
            if tmp_list:
   	         removedList = removedList + tmp_list

    # yeah, this is ugly, but look ma! lamdba/map magic!
    # make sure the lists are uniq, and make sure we remove the right packages
    filteredRemovedList = map(lambda x: x[0], removedList)
    blip = map(lambda x: ("%s-%s-%s-%s" % (x[0][0],x[0][1],x[0][2],x[0][3]), x) , removedList)

    d = {}
    for i in blip: d[i[0]] = i
    temp = d.values()

    filteredRemovedList = map(lambda x: x[1][0], temp)
    removedList = map(lambda x: x[1], temp)

    for pkg in filteredRemovedList:
	if pkg in pkgList:
            pkgList.remove(pkg)

    return removedList

# return a list of packages that need to be updated according
# to the preferences set in the config file, etc.
def getUpdatedPackageList(pkgNames, availList, fromDaemon, msgCallback = None,
                          progressCallback = None, refreshCallback = None):
    pkgSkipList = []

    if msgCallback != None:
        msgCallback(_("Retrieving list of all available packages"))

    if not len(pkgNames):
        # dont fetch a listall...

        if availList:
            pkgList = availList
        else:
            pkgList = getAvailablePackageList(msgCallback,progressCallback)

        pkgList = removeInstalledPackagesFromList(pkgList, [],
                                                  msgCallback,
                                                  progressCallback)
    else:
        # need a new xmlrpc call to just get info on the packages specified
        # it could reduce bandwidth a bit in exchange for more dynamic
        # content serverside
        if availList:
            pkgList = availList
        else:
            pkgList = getAvailablePackageList(msgCallback, progressCallback)

        # expand the vre less pkgNames (from the commandline)to have 0-0:0
        if not fromDaemon:
            pkgNames = map(lambda a: [a, '0','0',''], pkgNames)

	pkgList = removeInstalledPackagesFromList(pkgList, pkgNames,
                                                  msgCallback,
                                                  progressCallback)

        pkgList = removeNonPkgNamesPackagesFromList(pkgList,pkgNames,
                                                    msgCallback,
                                                    progressCallback)

    if len(pkgList) > 0:
        pkgSkipList = removeSkipPackagesFromList(pkgList, msgCallback,
                                                 progressCallback)

    if len(pkgList) > 0:
        if msgCallback != None:
            msgCallback(_("Getting headers for available packages"))
        total = len(pkgList)
        
        h_list = getHeaders(pkgList, msgCallback, progressCallback,
                            refreshCallback)

    if len(pkgList) > 0:
        pkgSkipList = pkgSkipList + removeSkipFilesPackagesFromList(pkgList, msgCallback,
                                                                    progressCallback)
            
    if len(pkgSkipList) > 0:
        if msgCallback != None:
            msgCallback(_("Getting headers for skipped packages"))
        total = len(pkgSkipList)

        for index in range(total):
            hdr = getHeader(pkgSkipList[index][0])
#            headerCache[pkgToStringArch(pkgSkipList[index][0])] = hdr
            if progressCallback:
                progressCallback(index+1, total)

    pkgList.sort()
    pkgSkipList.sort()
    return (pkgList, pkgSkipList)

    
def freeDiskSpace():
    import statvfs

    dfInfo = os.statvfs(cfg.readEntry("storageDir"))
    return long(dfInfo[statvfs.F_BAVAIL]) * (dfInfo[statvfs.F_BSIZE])


def rpmCallback(what, amount, total, key, cb):
    if what == rpm.RPMCALLBACK_INST_OPEN_FILE:
        pass
        #fd = os.open(key, os.O_RDONLY)
        #return fd
    elif what == rpm.RPMCALLBACK_INST_START:
        pass
    elif what == rpm.RPMCALLBACK_INST_CLOSE_FILE:
        print
    elif what == rpm.RPMCALLBACK_TRANS_PROGRESS:
        if cb:
            cb(amount, total)
        else:
            print "transaction %.5s%% done\r" % ((float(amount) / total) * 100),
    elif what == rpm.RPMCALLBACK_INST_PROGRESS:
        print "installation %.5s%% done\r" % ((float(amount) / total) * 100), 


def showSkipListHits(skiplist):
    errmsg = _("""
    To solve all dependencies for the RPMs you have selected, The following
    packages you have marked to exclude would have to be added to the set:
    """)
    errmsg = errmsg + _("\n    %-30.30s\t%-30.30s\n") % (
        _("Package Name"), _("Reason For Skipping"))
    errmsg = errmsg + " " * 4 +  "=" * 70 + "\n"
    for i in skiplist:
        package = "%s-%s-%s" % (i[0][0], i[0][1], i[0][2])
        errmsg = errmsg + "    %-30.30s\t%-30.30s\n" % (package, i[1])
    return errmsg

class DependencySolver:
    """This class is the engine of the dependency solver so that the external
       interface to up2date.dryRun remains readable"""
    def __init__(self, selected):
        # turn on debugging for this particular class here
        self.debug = 0 
        self.selected = selected
        available = getAvailablePackageList()

        # exclude the already selected ones from the available list
        xlist = map(lambda a: a[0], self.selected)
        self.available = {} 

        # a list of excludes that might break our dependency chains
        self.excludes = []

        # make sure the ones that are skipped are not in here
        available, junk = self.__skip(available)
        for i in range(len(available)):
	    pkgName = available[i][0]
            if not pkgName in xlist:
		if not self.available.has_key(pkgName):
		    self.available[pkgName] = available[i]

        # placeholder for the database and transaction set
        self.db = self.ts = None
        
        # null callbacks for now
        self.msgCallback = self.progressCallback = self.refreshCallback = None
        
    # set the callbacks
    def set_callback(self, name, function):
        if name == "msg":
            self.msgCallback = function
        elif name == "progress":
            self.progressCallback = function
        elif name == "refresh":
            self.refreshCallback = function

    # get ready to rumble
    def setup(self):
        # prepare for the transaction list
        self.db = openrpmdb()
        self.ts = rpm.TransactionSet("/", self.db)

        if self.msgCallback != None:
            self.msgCallback(_("Testing package set / solving RPM inter-dependencies"))
        pkgset = self.selected
        self.selected = [] # the ___add will fill it back as it does ts.add
        self.__add(pkgset, check = 0)
        self.ts.order()

    # refresh the screen or something
    def __refresh(self):
        if self.refreshCallback:
            self.refreshCallback()
        
    # sanitize a list of packages for the ones we have to skip
    def __skip(self, plist):
        # make sure the new packages don't include any packages
        # with packages excludes
        plist.sort()
        xlist = removeSkipPackagesFromList(plist)
        for x in xlist:
            # remove the package
            if self.debug:
                print "Removing package", x
	    pkgName = x[0][0]
            if self.available.has_key(pkgName): # make it unavailable too
		del self.available[pkgName]
            self.excludes.append(x)
        return plist, xlist

    # add packages to a transaction set
    def __add(self, plist, check = 0):
        if self.debug:
            print "Candidates for the selected list:", plist
        # get the headers
        h_list = getHeaders(plist, refreshCallback = self.refreshCallback)

        added = [] # what packages do we add?
        
        # add them one by one
        counter = -1
        for h in h_list:
            counter = counter + 1
            if not h:
                continue
            if check:
                xlist = checkHeaderForFileConfigExcludes(h, plist[counter], self.db)
                if xlist: # those packages have to be skipped too
                    if self.debug:
                        print "Got exclude list from config file check:", xlist
                    for x in xlist:
                        if self.debug:
                            print "Skipping package %s because of config file change" % (x,)
			pkgName = x[0][0]
                        if self.available.has_key(pkgName):
                            del self.available[pkgName]
                        self.excludes.append(x)
                    continue
            newp = plist[counter]
            if newp not in self.selected:
                self.selected.append(newp)
                if self.debug:
                    print "Adding %s to transaction set" % newp
                self.ts.add(h, newp, "u")
                added.append(newp)
            if self.available.has_key(newp[0]): # no longer available
                del self.available[newp[0]]
            self.__refresh()
        return added
    
    # Solve dependencies
    def __dependencies(self, dependencies):
        if not dependencies:
            return 0
        changed = 0
        if self.debug:
            print "Dependencies:", dependencies
        solved = solveDep(map(lambda a: a[1], dependencies),
                          refreshCallback = self.refreshCallback)
        if self.debug:
            print "Got back response:", solved
        # If we got a response back, take out the packages that have been
        # already selected so we don't loop without a purpose
        for s in solved[:]:
            if s in self.selected:
                if self.debug:
                    print "Already selected:", s
                solved.remove(s)
        # if we did not get a solution pack back, try looking at the
        # problem packages that are thorwing these dependencies.
        if not solved:
            solved = []
            snames = map(lambda a: a[0], dependencies)
	    for s in snames:
	    	if self.available.has_key(s):
                    if self.debug:
                        print "Package %s raised invalid dependency. Adding %s to set" % (
                        s, self.available[s])
                    solved.append(self.available[s])
                    changed = 1
        # sanitize this list
        solved, skiplist = self.__skip(solved)
        added = self.__add(solved, check = 1)
        if added:
            changed = 1
        return changed
        
    # Solve conflicts
    def __conflicts(self, conflicts):        
        if not conflicts:
            return 0
        changed = 0
        if self.debug:
            print "Conflicts:", conflicts
        solved = []
        for c in conflicts:
            # c should be a pair (name1, name2)
            # if one of these packages is still available, grab it
            for conflname in c:
                if self.available.has_key(conflname):
                    solved.append(self.available[conflname])
                    changed = 1
        if solved is None and not changed:
            raise DependencyError(_("Could not satisfy conflict dependencies."))
        solved, skiplist = self.__skip(solved)
        added = self.__add(solved, check = 1)
        if added:
            changed = 1
        return changed
    
    # Format the dependency error nicely for the plain guy
    def __dep_error(self, deps):
        s = StringIO.StringIO()
        for dep in deps:
            ((name, version, release), (needsName, needsVersion),
             flags, suggested, sense) = dep
            if needsVersion:
                s.write(_("%-40.40s requires %s\n") % (
                    "%s-%s-%s" % (name, version, release),
                    "%s-%s" % (needsName, needsVersion)))
            else:
                s.write(_("%-40.40s requires %s\n") % (
                    "%s-%s-%s" % (name, version, release),
                    needsName))
        ret = s.getvalue()
        s.close()
        return ret


    def __filter_excludes_for_deps(self,deps):
        """This just filters the exclude list to any that are
        actually applicable to the failed deps"""
        # FIXME: this is probabaly fragile
        # FIXME: return the applicable and the non applicable
        #        in seperate list so we can display them as well?

        depnames = []
        for dep in deps:
            ((name, version, release), (needsName, needsVersion),
             flags, suggested, sense) = dep
            depnames.append(name)

        new_excludes = []
        for exclude in self.excludes:
            if exclude[0][0] in depnames:
                new_excludes.append(exclude)

        return new_excludes
            
    # generic dependency handler
    def process_deps(self, deps):
        if not deps:
            return 0
        conflicts = []
        dependencies = []
        last_excludes = self.excludes
        self.excludes = []
        for dep in deps:
            if self.debug:
                print "Processing dependency", dep
            ((name, version, release), (needsName, needsVersion),
             flags, suggested, sense) = dep
            if sense == rpm.RPMDEP_SENSE_REQUIRES:
                dependencies.append(name, needsName)
            elif sense == rpm.RPMDEP_SENSE_CONFLICTS:
                conflicts.append(name, needsName)
            else:
                raise DependencyError(_("Don't know how to handle dependency sense \"%s\"") %
                                      sense)
            self.__refresh()
        # now solve the dependencies...
        changed = self.__dependencies(dependencies)
        self.__refresh()
        # and the conflicts...
        changed = changed + self.__conflicts(conflicts)
        self.__refresh()
        if not changed: # blow up
            if not self.excludes: # get the last round instead
                self.excludes = last_excludes
            errmsg = ""
            applicable_excludes = self.__filter_excludes_for_deps(deps) 
            if len(applicable_excludes):
                errmsg = showSkipListHits(self.excludes)
            dep_error = self.__dep_error(deps)
            raise DependencyError(_("%s\n==Unresolvable chain of dependencies:\n"
                                        "%s") % (errmsg, dep_error))
        return 1
    
    # run a dependency iteration
    def solvedep(self):
        if self.debug:
            print "\nChecking for dependencies"        
        self.__refresh()
        deps = self.ts.depcheck()       
        if self.debug and deps:
            print "RPM returned %d deps." % len(deps)
        if deps:
            # make the deps list unique:
            deps.sort()
            last = deps[-1]
            for i in range(len(deps)-2, -1, -1):
                if last == deps[i]:
                    del deps[i]
                else:
                    last = deps[i]
        ret = self.process_deps(deps)
        return ret

    # run the transaction set now to check for any other problems
    def run(self):
        if self.debug:
            print "Running transaction (final step)..."
        self.ts.order()
        self.__refresh()
        ret = self.ts.run(rpm.RPMTRANS_FLAG_BUILD_PROBS, 0,
                          rpmCallback, self.progressCallback)
        self.__refresh()
        return ret

# The new dryRun
def dryRun(pkgList, msgCallback = None, progressCallback = None, refreshCallback=None):
    # XXX: initialize correctly from the --verbose cmd line flag.
    debug = 0
    
    if debug:
        print "Called dryRun:", pkgList

    depsolve = DependencySolver(pkgList)

    # set the callbacks wwe need
    depsolve.set_callback("msg", msgCallback)
    depsolve.set_callback("progress", progressCallback)
    depsolve.set_callback("refresh", refreshCallback)
    
    # set up the trabsaction set & all
    depsolve.setup()
    while 1:
        if debug:
            print "\nSelected so far", depsolve.selected
        ret = depsolve.solvedep()
        if debug:
            print "Got code %d from solvedep" % ret
        if ret == 0:
            break
        elif ret < 0: # not really reached
            raise DependencyError(_("Could not complete dependency resolution"))
    rc = depsolve.run()
    if progressCallback:
        progressCallback(1, 1)       
    if len(rc):
        msg = ""
        for code in rc:
            msg = msg + code[0] + "\n"
            raise ConflictError(_("Test install failed because of package conflicts:\n") + msg)
    newList = []
    for s in depsolve.selected:
        if not s in pkgList:
            pkgList.append(s)
            newList.append(s)
    return (newList, [])

def pprint_pkglist(pkglist):
    if type(pkglist) == type([]):
        foo = map(lambda a : "%s-%s-%s" % (a[0],a[1],a[2]), pkglist)
    else:
        foo = "%s-%s-%s" % (pkglist[0], pkglist[1], pkglist[2])
    return foo
                 
def installPackages(pkgList, rpmCallback):
    try:
        log.log_me("installing packages: %s " % pprint_pkglist(pkgList))
    except:
        pass

    # a list of all the packge names to treat as kernels.
    # "hardcode" it here so people dont mess with it as much
    total = len(pkgList)

    oldInstalledPackages = getInstalledPackageList()
    
    # delete any existing db, open in write mode
    db = openrpmdb(option=1)
    ts = rpm.TransactionSet("/", db)

    kernels = []
    for p_i in range(total):
        hdr = getHeader(pkgList[p_i])
        if hdr:
            if "kernel" in hdr['provides']:
                ts.add(hdr,hdr, "i")
                kernels.append(hdr)
            else:
                ts.add(hdr, hdr, "u")
        else:
            raise RpmError(_("Could not find header for %s") % pkgList[p_i])

    ts.order()
    deps = ts.depcheck()
    if deps:
        raise DependencyError(_("Dependencies should have already been resolved, but they are not."))

    if cfg.readEntry("justdb"):
        rc = ts.run(rpm.RPMTRANS_FLAG_JUSTDB, 0,
                    rpmCallback, cfg.readEntry("storageDir"))
    else:
        rc = ts.run(0, 0, rpmCallback, cfg.readEntry("storageDir"))
    if rc:
        errors = "\n"
        for e in rc:
            try:
                errors = errors + e[1] + "\n"
            except:
                errors = errors + str(e) + "\n"
        raise RpmError(_("Failed installing packages: %s") % errors)

    newInstalledPackages = getInstalledPackageList()

    removedPackages = []
    addedPackages = []

    # find packages that were removed
    for pkg in oldInstalledPackages:
        if pkg not in newInstalledPackages:
            removedPackages.append(pkg)

    for pkg in newInstalledPackages:
        if pkg not in oldInstalledPackages:
            addedPackages.append(pkg)

    if len(removedPackages):
    	remoteDelPackages(removedPackages)
    if len(addedPackages):    
    	remoteAddPackages(addedPackages)

    return kernels

def removePackages(pkgList, rpmCallback):
    try:
        log.log_me("removing packages: %s " % pprint_pkglist(pkgList))
    except:
        pass  

    total = len(pkgList)

    oldInstalledPackages = getInstalledPackageList()

    # delete any existing db, open in write mode
    db = openrpmdb(option=1)
    ts = rpm.TransactionSet("/", db)

    # FIXME: just that a package is installed before claiming
    # it cant be removed

    remove_list = []
    for p_i in range(total):
        for pattern in cfg.readEntry("removeSkipList"):
            pkgName = "%s-%s-%s" % (pkgList[p_i][0],pkgList[p_i][1],pkgList[p_i][2])
            # check the package name against the globs in removeSkipList
            if fnmatch.fnmatch(pkgList[p_i][0], pattern):
                raise RpmRemoveSkipListError, pkgName
            else:
                remove_list.append(pkgName)

    failed_packages = {}
    for pkgName in remove_list:
        try:
            ts.remove(pkgName)
        except rpm.error, e:
            failed_packages[pkgName] = e

    # if we get failed packages, raise the hash
    # of failed packages
    if len(failed_packages.keys()):
        raise RpmRemoveError, failed_packages
    
    ts.order()
    deps = ts.depcheck()
    
    # FIXME: return failed deps if any
    if deps:
        raise DependencyError, deps

    if cfg.readEntry("justdb"):
        # do we use storage dir here?? do we need it at all?
        rc = ts.run(rpm.RPMTRANS_FLAG_JUSTDB, 0,
                    rpmCallback, cfg.readEntry("storageDir"))
    else:
        rc = ts.run(0, 0, rpmCallback, cfg.readEntry("storageDir"))
    if rc:
        errors = "\n"
        for e in rc:
            try:
                errors = errors + e[1] + "\n"
            except:
                errors = errors + str(e) + "\n"
        raise RpmError(_("Failed removing packages: %s") % errors)

    newInstalledPackages = getInstalledPackageList()
    removedPackages = []
    # find packages that were removed

    for pkg in oldInstalledPackages:
        if pkg not in newInstalledPackages:
            removedPackages.append(pkg)

    if len(removedPackages):
    	remoteDelPackages(removedPackages)
        log.log_me("Removed packages: %s" % removedPackages)

    
    return 0

def installBootLoader(kernels):
    
    # option to disable this permanently for those using grub,etc
    if cfg.readEntry("noBootLoader"):
        return

    # dont run the new lilo if were running "justdb"
    if cfg.readEntry("justdb"):
        return

    log.log_me("Modifying bootloader config to include the new kernel info")
    kernelList = []
    # annoying, but kernels provide "kernel", "kernel-drm", "kernel-dfasdf", etc 
    for krn in kernels:
        if "kernel-BOOT" in krn['provides']:
            extraInfo = "kernel-BOOT"
        elif "kernel-smp" in krn['provides']:
            extraInfo = "kernel-smp"
        elif "kernel-enterprise" in krn['provides']:
            extraInfo = "kernel-enterprise"
        else:
            extraInfo = "kernel"

        # this logics a bit weird
        if extraInfo == None:
            infoString = ""
        elif extraInfo == "kernel":
            infoString = ""
        elif extraInfo == "kernel-BOOT":
            infoString = "BOOT"
        elif extraInfo == "kernel-enterprise":
            infoString = "enterprise"
        elif extraInfo == "kernel-smp":
            infoString = "smp"
        else:
            infoString = ""


        verRel = "%s-%s%s" % (krn['version'], krn['release'],infoString)
        kernelList.append((verRel,extraInfo))
        log.log_me("Adding %s to bootloader config" % verRel)


    bootloader = checkbootloader.whichBootLoader()
    
    if up2dateUtils.archGlob() == 'i?86' and bootloader == "LILO":
        __install_lilo(kernelList)
    elif up2dateUtils.archGlob() == 'i?86' and bootloader == "GRUB":
        # at the moment, kernel rpms are supposed to grok grub
        # and do the Right Thing. Just a placeholder for doc purposes
        __install_grub(kernelList)
    elif up2dateUtils.archGlob() == "ia64":
        __install_elilo(kernelList)

def __install_elilo(kernelList):
    import elilocfg
        
    log.log_me("Running elilo with the new configuration")
    ret = elilocfg.installNewImages(kernelList,test=0)
    return ret

def __install_lilo(kernelList):
    import lilocfg
    
    log.log_me("Running lilo with the new configuration")
    ret = lilocfg.installNewImages(kernelList,test=0)
    return ret

def __install_grub(kernelList):
    import grubcfg

    log.log_me("Installing the kernel via grub")
    ret  = grubcfg.installNewImages(kernelList,test=0)
    return ret

                        
def stdoutMsgCallback(msg):
    print msg

# set a global message callback for the module
# default to the stdout one
def setWarningCallback(callback):
    global warningCallback
    warningCallback = callback

warningCallback = stdoutMsgCallback
global warningCallback

def simpleCallback(count, total):
    print "%.5s%% done\r" % ((float(count) / total) * 100),

